(* ::Package:: *)

EscollirPunt::usage="EscollirPunt[{a_,b_,p_}] Escull un punt a l'atzar de E(Fp).";
ComprovarPunt::usage="ComprovarPunt[{x_,y_},{a_,b_,p_}] Comprova si el punt \[EAcute]s de E(Fp).";
SumaPunts::usage="SumaPunts[{x_,y_},{w_,z_}]Suma dos punts a E(Fp) amb E: y^2=x^3+ax+b.";
MultiplicaPunt::usage="MultiplicaPunt[k_,{x_,y_},{a_,b_,p_}]Multiplica un punt de E(Fp) per un enter K.";
DeterminarOrdrePunt::usage="DeterminarOrdrePunt[{x_,y_},{a_,b_,p_}] Determina l'ordre d'un punt a for\[CCedilla]a bruta."
BabyGiant::usage="BabyGiant[{x_,y_},{a_,b_,p_}] Determina l'ordre d'un punt mitjan\[CCedilla]ant el m\[EGrave]tode Baby Step, Giant Step."
DeterminarOrdreCorba::usage="DeterminarOrdreCorba[{a_,b_,p_}] Determina l'orde de la corba mitjan\[CCedilla]ant el m\[EGrave]tode Baby Step, Giant Step."
ComprovarCorba::usage="ComprovarCorba[{a_,b_,p_}] Comprova si la corba est\[AGrave] ben definida.";
EscollirCorbaCr::usage="[] Escull aleat\[OGrave]riament una corba \[UAcute]til per encriptar."
EscollirPuntCr::usage="EscollirPuntCr[{a_,b_,p_}] Escull aleat\[OGrave]riament un punt \[UAcute]til per encriptar, \[EAcute]s a dir, d'ordre elevat."
EnterAPunt::usage="EnterAPunt[m_,{a_,b_,p_}] Converteix un enter m<(p/100-100) en un punt de la corba."
PuntAEnter::usage="PuntAEnter[{x_,y_}] Retorna un enter d'un punt generat per la funci\[OAcute] EnterAPunt."

Begin["corbeseliptiques`"];

EscollirPunt[{a_,b_,p_}]:=Flatten[{
	resc=RandomInteger[{0,p-1}];
	sesc=Mod[resc^3+a*resc+b,p];
	If[
		JacobiSymbol[sesc,p]==1,
		{resc,PowerMod[sesc,1/2,p]},
		EscollirPunt[{a,b,p}]
	]
}];

ComprovarPunt[{x_,y_},{a_,b_,p_}]:=
	If[
		x==\[Infinity],
		"True",
		If[
			Mod[y^2,p]==Mod[x^3+a*x+b,p],
			"True",
			"False"
		]
	]
	
SumaPunts[{x_,y_},{w_,z_},{a_,b_,p_}]:=
	If[x==\[Infinity]||w==\[Infinity],
		If[
			x==\[Infinity],
			{w,z},
			{x,y}
		],
		If[
			x!=w,
				msum=(z-y)*PowerMod[w-x,-1,p];
				rsum=Mod[(msum)^2-x-w,p];
				ssum=Mod[(msum)(x-rsum)-y,p]; 
			{rsum,ssum},
			If[
				y!=z,
				{\[Infinity],\[Infinity]},
				If[
					y!=0,
					msum=(3x^2+a)*PowerMod[2y,-1,p];
					rsum=Mod[msum^2-2x,p];
					ssum=Mod[msum(x-rsum)-y,p];
					{rsum,ssum},
					{\[Infinity],\[Infinity]}
				]
			]		
		]
	]
	
MultiplicaPunt[k_,{x_,y_},{a_,b_,p_}]:=Flatten[{
	If[
		k==0,
		{\[Infinity],\[Infinity]},
		If[k>0, nmul=k, nmul=-k];
		vmul={};
		imul=0;
		While[
			nmul!=1,
			lmul=QuotientRemainder[nmul,2];
			nmul=lmul[[1]];
			vmul=Append[vmul,lmul[[2]]];
			imul+=1;
		];
		If[k>0, {rmul,smul}={x,y}, {rmul,smul}={x,-y}];
		While[
			imul>0,
			{rmul,smul}=SumaPunts[{rmul,smul},{rmul,smul},{a,b,p}];
			If[vmul[[imul]]==1, {rmul,smul}=SumaPunts[{rmul,smul},{x,y},{a,b,p}]; ];
			imul-=1;
		];
		{rmul,smul}
	]
}];

DeterminarOrdrePunt[{x_,y_},{a_,b_,p_}]:={
	{rdet,sdet}={x,y};
	idet=1;
	While[
		rdet!=\[Infinity],
		{rdet,sdet}=SumaPunts[{rdet,sdet},{x,y},{a,b,p}];
		idet+=1;
	];
	idet
}[[1]]

BabyGiant[{x_,y_},{a_,b_,p_}]:={
	qbg=MultiplicaPunt[p+1,{x,y},{a,b,p}];
	mbg=IntegerPart[p^(1/4)]+1;
	jbg=1;
	lbg={\[Infinity]};
	{rbg,sbg}={\[Infinity],\[Infinity]};
	While[
		jbg<=mbg,
		{rbg,sbg}=SumaPunts[{rbg,sbg},{x,y},{a,b,p}];
		lbg=Append[lbg,rbg];
		jbg+=1;
	]; (* {r,s}=mP *)
	auxbg=MultiplicaPunt[2,{rbg,sbg},{a,b,p}]; (* aux=2mP *)
	{rbg,sbg}=MultiplicaPunt[mbg,auxbg,{a,b,p}]; (* {r,s}=m2mP *)
	{rbg,sbg}={rbg,-sbg}; (* {r,s}=-m2mP *)
	condbg=0;
	ibg=-mbg;
	jbg=0;
	{rbg,sbg}=SumaPunts[qbg,{rbg,sbg},{a,b,p}]; (* {r,s}=Q-m2mP *)
	While[
		jbg<(mbg+1) && condbg==0,
		jbg+=1;
		If[
			rbg==lbg[[jbg]],
			condbg=1;
		];
	];
	While[
		condbg==0 && ibg<mbg,
		ibg+=1;
		{rbg,sbg}=SumaPunts[{rbg,sbg},auxbg,{a,b,p}];
		jbg=0;
		While[
			jbg<(mbg+1) && condbg==0,
			jbg+=1;
			If[
				rbg==lbg[[jbg]],
				condbg=1;
			];
		];
	];
	jbg-=1; (*lbg[[1]] es correspon amb 0P *)
	auxbg=MultiplicaPunt[jbg,{x,y},{a,b,p}];
	If[auxbg[[2]]!=sbg, jbg=-jbg];
	ordbg=p+1+2*mbg*ibg-jbg; (*En aquest moment tenim Q+i2mP=jP*)
	factorbg=FactorInteger[ordbg];
	ibg=Dimensions[factorbg][[1]];
	jbg=1;
	While[
		jbg<=ibg,
		If[
			MultiplicaPunt[ordbg/factorbg[[jbg]][[1]],{x,y},{a,b,p}][[1]]==\[Infinity],
			ordbg=ordbg/factorbg[[jbg]][[1]];
			kbg=1;
			While[
				kbg<factorbg[[jbg]][[2]] && MultiplicaPunt[ordbg/factorbg[[jbg]][[1]],{x,y},{a,b,p}][[1]]==\[Infinity],
				ordbg=ordbg/factorbg[[jbg]][[1]];
				kbg+=1;
			];
		];
		jbg+=1;
	];
	ordbg
}[[1]]

DeterminarOrdreCorba[{a_,b_,p_}]:={
	odoc=1;
	mdoc=1;
	idoc=(4*p^(1/2)+1);
	While[
		odoc<idoc,
		{xdoc,ydoc}=EscollirPunt[{a,b,p}];
		odoc=LCM[odoc,BabyGiant[{xdoc,ydoc},{a,b,p}]];
	];
	mdoc=odoc;
	While[
		mdoc<p+1-2*p^(1/2),
		mdoc+=odoc;
	];
	mdoc
}[[1]]

ComprovarCorba[{a_,b_,p_}]:= {
auxcom=DeterminarOrdreCorba[{a,b,p}];
factorcom=FactorInteger[BabyGiant[EscollirPunt[{a,b,p}],{a,b,p}]];
If[
	Mod[4a^3+27b^2,p]!=0 
	&& auxcom!=p
	&& auxcom!=p+1
	&& factorcom[[Dimensions[factorcom][[1]]]][[1]]>(p/100)
	, 1,0]
}[[1]]

EscollirCorbaCr:={
	aecr=RandomInteger[{-1000000000,1000000000}];
	becr=RandomInteger[{-1000000000,1000000000}];
	pecr=Prime[RandomInteger[{900000000,1000000000}]];
	While[
		ComprovarCorba[{aecr,becr,pecr}]==0,
		aecr=RandomInteger[{-1000000000,1000000000}];
		becr=RandomInteger[{-1000000000,1000000000}];
		pecr=Prime[RandomInteger[{900000000,1000000000}]];
	];
	aecr,becr,pecr
}

EscollirPuntCr[{a_,b_,p_}]:=Flatten[{
	rescr=RandomInteger[{0,p-1}];
	sescr=Mod[rescr^3+a*rescr+b,p];
	If[
		JacobiSymbol[sescr,p]==1 ,
		sescr=PowerMod[sescr,1/2,p];
		ordescr=BabyGiant[{rescr,sescr},{a,b,p}];	
		If[
			ordescr>1000000000,
			{rescr,sescr},
			EscollirPuntCr[{a,b,p}]
		],
		EscollirPuntCr[{a,b,p}]
	]
}];

EnterAPunt[m_,{a_,b_,p_}]:=
	If[
		IntegerQ[m] && m<(p/100-100),
		xeap=100m;
		seap=Mod[xeap^3+a*xeap+b,p];
		ieap=0;
		While[
			JacobiSymbol[seap,p]!=1 && ieap<100,
			xeap+=1;
			seap=Mod[xeap^3+xeap*a+b,p];
			ieap+=1;
		];
		If[
			JacobiSymbol[seap,p]==1,
			{xeap,PowerMod[seap,1/2,p]},
			"No s'ha pogut expressar el missatge com a punt de la corba"
		],
	"Missatge no v\[AGrave]lid"
]

PuntAEnter[{x_,y_}]:= IntegerPart[x/100]


		
End[];






